"""This is a Wind River Linux driver implementation."""

from condoor.drivers.generic import Driver as Generic
from condoor import pattern_manager


class Driver(Generic):
    """This is a Driver class implementation for Wind River Linux."""

    platform = 'Windriver'
    inventory_cmd = None
    target_prompt_components = ['prompt_dynamic', 'prompt_default', 'calvados', 'lc']
    prepare_terminal_session = []
    families = {
    }

    def __init__(self, device):
        """Initialize the Wind River Linux driver object."""
        super(Driver, self).__init__(device)

    def get_version_text(self):
        """Return the version information."""
        version_text = self.device.send('cat /etc/issue', timeout=10)
        return version_text

    def update_driver(self, prompt):
        """Return driver name based on prompt analysis."""
        return pattern_manager.platform(prompt, ['Windriver', 'Calvados', 'eXR'])

    def get_os_type(self, version_text):
        """Return Windriver os type."""
        return 'Windriver'



"""This is a QNX driver implementation."""

from condoor.drivers.generic import Driver as Generic
from condoor import pattern_manager


class Driver(Generic):
    """This is a Driver class implementation for Wind River Linux."""

    platform = 'QNX'
    inventory_cmd = None
    target_prompt_components = ['prompt_dynamic', 'prompt_default', 'xr']
    prepare_terminal_session = []
    families = {
    }

    def __init__(self, device):
        """Initialize the QNX Linux driver object."""
        super(Driver, self).__init__(device)

    def get_version_text(self):
        """Return the version information from Unix host."""
        version_text = self.device.send('uname -sr', timeout=10)
        return version_text

    def update_driver(self, prompt):
        """Return driver name based on prompt analysis."""
        return pattern_manager.platform(prompt, ['QNX', 'XR'])

    def get_os_type(self, version_text):
        """Return QNX os type."""
        return 'QNX'



"""This is IOS XR 64 bit driver implementation."""

from functools import partial
import re
import pexpect

from condoor.exceptions import CommandSyntaxError, CommandTimeoutError, ConnectionError, ConnectionAuthenticationError
from condoor.actions import a_connection_closed, a_expected_prompt, a_stays_connected, a_unexpected_prompt, a_send, \
    a_store_cmd_result, a_message_callback, a_send_line, a_reconnect, a_send_boot, a_return_and_reconnect
from condoor.utils import pattern_to_str
from condoor.fsm import FSM
from condoor.drivers.XR import Driver as XRDriver
from condoor import pattern_manager, EOF
from condoor.config import CONF

_C = CONF['driver']['eXR']


class Driver(XRDriver):
    """This is a Driver class implementation for IOS XR 64 bit."""

    platform = 'eXR'
    target_prompt_components = ['prompt_dynamic', 'prompt_default', 'rommon', 'xml']
    reload_cmd = 'admin hw-module location all reload'
    families = {
        "ASR9K": "ASR9K",
        "ASR-9": "ASR9K",
        "ASR9": "ASR9K",
        "NCS-6": "NCS6K",
        "NCS-4": "NCS4K",
        "NCS-50": "NCS5K",
        "NCS-55": "NCS5500",
        "NCS-54": "NCS540",
        "NCS-56": "NCS560",
        "NCS-1001": "NCS1001",
        "NCS-1": "NCS1K",
        "NCS1": "NCS1K",
        "IOS-XRv 9000": "IOSXRv-9K",
        "IOS XRv x64": "IOSXRv-X64",
    }

    def __init__(self, device):
        """Initialize the XR 64 bit Driver object."""
        super(Driver, self).__init__(device)
        self.calvados_re = pattern_manager.pattern(self.platform, 'calvados')
        self.calvados_connect_re = pattern_manager.pattern(self.platform, 'calvados_connect')
        self.calvados_term_length = pattern_manager.pattern(self.platform, 'calvados_term_length')

    def update_driver(self, prompt):
        """Return driver name based on prompt analysis."""
        return pattern_manager.platform(prompt, ['eXR', 'Calvados', 'Windriver'])

    def get_version_text(self):
        """Return version information text."""
        version_text = self.device.send("show version", timeout=120)
        return version_text

    def wait_for_string(self, expected_string, timeout=60):
        """Wait for string FSM for XR 64 bit."""
        # Big thanks to calvados developers for make this FSM such complex ;-)
        #                    0                         1                        2                        3
        events = [self.syntax_error_re, self.connection_closed_re, expected_string, self.press_return_re,
                  #        4           5                 6                7               8
                  self.more_re, pexpect.TIMEOUT, pexpect.EOF, self.calvados_re, self.calvados_connect_re,
                  #     9
                  self.calvados_term_length]

        # add detected prompts chain
        events += self.device.get_previous_prompts()  # without target prompt

        self.log("Expecting: {}".format(pattern_to_str(expected_string)))
        self.log("Calvados prompt: {}".format(pattern_to_str(self.calvados_re)))

        transitions = [
            (self.syntax_error_re, [0], -1, CommandSyntaxError("Command unknown", self.device.hostname), 0),
            (self.connection_closed_re, [0], 1, a_connection_closed, 10),
            (pexpect.TIMEOUT, [0, 2], -1, CommandTimeoutError("Timeout waiting for prompt", self.device.hostname), 0),
            (pexpect.EOF, [0, 1], -1, ConnectionError("Unexpected device disconnect", self.device.hostname), 0),
            (self.more_re, [0], 0, partial(a_send, " "), 10),
            (expected_string, [0, 1], -1, a_expected_prompt, 0),
            (self.calvados_re, [0], -1, a_expected_prompt, 0),
            (self.press_return_re, [0], -1, a_stays_connected, 0),
            (self.calvados_connect_re, [0], 2, None, 0),
            # admin command to switch to calvados
            (self.calvados_re, [2], 3, None, _C['calvados_term_wait_time']),
            # getting the prompt only
            (pexpect.TIMEOUT, [3], 0, partial(a_send, "\r\r"), timeout),
            # term len
            (self.calvados_term_length, [3], 4, None, 0),
            # ignore for command start
            (self.calvados_re, [4], 5, None, 0),
            # ignore for command start
            (self.calvados_re, [5], 0, a_store_cmd_result, 0),
        ]

        for prompt in self.device.get_previous_prompts():
            transitions.append((prompt, [0, 1], 0, a_unexpected_prompt, 0))

        fsm = FSM("WAIT-4-STRING", self.device, events, transitions, timeout=timeout)
        return fsm.run()

    def reload(self, reload_timeout, save_config):
        """Reload the device."""
        MAX_BOOT_TIME = 1800  # 30 minutes - TODO(klstanie): move to config
        RELOAD_PROMPT = re.compile(re.escape("Reload hardware module ? [no,yes]"))

        START_TO_BACKUP = re.compile("Status report.*START TO BACKUP")
        BACKUP_IN_PROGRESS = re.compile("Status report.*BACKUP INPROGRESS")

        BACKUP_HAS_COMPLETED_SUCCESSFULLY = re.compile("Status report.*BACKUP HAS COMPLETED SUCCESSFULLY")
        DONE = re.compile(re.escape("[Done]"))

        STAND_BY = re.compile("Please stand by while rebooting the system")
        CONSOLE = re.compile("ios con[0|1]/(?:RS?P)?[0-1]/CPU0 is now available")
        CONFIGURATION_COMPLETED = re.compile("SYSTEM CONFIGURATION COMPLETED")
        CONFIGURATION_IN_PROCESS = re.compile("SYSTEM CONFIGURATION IN PROCESS")
        BOOTING = re.compile("Booting IOS-XR 64 bit Boot previously installed image")

        #           0                 1              2                         3                          4     5
        events = [RELOAD_PROMPT, START_TO_BACKUP, BACKUP_IN_PROGRESS, BACKUP_HAS_COMPLETED_SUCCESSFULLY, DONE, BOOTING,
                  #   6                  7               8                     9                         10
                  CONSOLE, self.press_return_re, CONFIGURATION_COMPLETED, CONFIGURATION_IN_PROCESS, self.username_re,
                  # 11          12           13           14
                  EOF, pexpect.TIMEOUT, self.rommon_re, STAND_BY]

        transitions = [
            # do I really need to clean the cmd
            (RELOAD_PROMPT, [0], 1, partial(a_send_line, "yes"), MAX_BOOT_TIME),
            (START_TO_BACKUP, [0, 1], 2, a_message_callback, 60),
            (BACKUP_IN_PROGRESS, [0, 1, 2], 2, a_message_callback, 60),
            (BACKUP_HAS_COMPLETED_SUCCESSFULLY, [0, 1, 2], 3, a_message_callback, 10),
            (DONE, [1, 2, 3], 4, None, MAX_BOOT_TIME),
            (STAND_BY, [2, 3, 4], 5, a_message_callback, MAX_BOOT_TIME),
            (self.rommon_re, [0, 4], 5, partial(a_send_boot, "boot"), MAX_BOOT_TIME),
            (BOOTING, [0, 1, 2, 3, 4], 5, a_message_callback, MAX_BOOT_TIME),
            (CONSOLE, [0, 1, 2, 3, 4, 5], 6, None, 600),
            (self.press_return_re, [6], 7, partial(a_send, "\r"), 300),
            (CONFIGURATION_IN_PROCESS, [7], 8, None, 180),
            (CONFIGURATION_COMPLETED, [8], -1, a_reconnect, 0),
            (self.username_re, [9], -1, a_return_and_reconnect, 0),
            (EOF, [0, 1, 2, 3, 4, 5], -1, ConnectionError("Device disconnected"), 0),
            (pexpect.TIMEOUT, [7], 9, partial(a_send, "\r"), 180),
            (pexpect.TIMEOUT, [1, 5, 8], -1, ConnectionError("Boot process took more than {}s".format(MAX_BOOT_TIME)), 0),
            (pexpect.TIMEOUT, [9], -1, ConnectionAuthenticationError("Unable to reconnect after reloading"), 0)
        ]

        fsm = FSM("RELOAD", self.device, events, transitions, timeout=600)
        return fsm.run()



"""Provides Device class representing the physical device for both target and jumphost."""

import sys
import pexpect

from condoor.exceptions import ConnectionError, CommandSyntaxError, CommandTimeoutError
from condoor.utils import parse_inventory
from condoor.fsm import FSM


class Device(object):
    """Device class representing physical device for both target and jumphost."""

    def __init__(self, chain, node_info, driver_name='jumphost', is_target=False):
        """Initialize Device object."""
        self.chain = chain
        self.hostname = "{}:{}".format(node_info.hostname, node_info.port)  # used by driver

        self.node_info = node_info
        self.ctrl = None

        # information whether the device is connected to the console
        self.is_console = False

        # True is last device in the chain
        self.is_target = is_target

        # prompt
        self.prompt = None
        self.prompt_re = None

        # properties with getter to collect in case None
        # version_text
        self._version_text = None

        # inventory_text
        self._inventory_text = None

        # hostname_text
        self._hostname_text = None

        # show users text
        self._users_text = None

        # set if device is connect
        self.connected = False

        self.mode = None

        self.protocol = None
        self.driver = self.make_driver(driver_name)

        self.os_version = None
        self.os_type = None
        self.family = None
        self.platform = None
        self.udi = None
        self.is_console = None

        self.last_command_result = None

        self.last_error_msg = None

    @property
    def device_info(self):
        """Return device info dict."""
        return {
            'family': self.family,
            'platform': self.platform,
            'os_type': self.os_type,
            'os_version': self.os_version,
            'udi': self.udi,
            # TODO(klstanie): add property to make driver automatically
            'driver_name': self.driver.platform,
            'mode': self.mode,
            'is_console': self.is_console,
            'is_target': self.is_target,
            # 'prompt': self.driver.base_prompt(self.prompt),
            'hostname': self.hostname,
        }

    @device_info.setter
    def device_info(self, info):
        for key, value in info.items():
            self.chain.connection.log("Update: [{}] {}<-{}".format(self, key, value))
            setattr(self, key, value)

    def __repr__(self):
        """Return string representing node info."""
        return str(self.node_info)

    def clear_info(self):
        """Clear the device info."""
        self._version_text = None
        self._inventory_text = None
        self._users_text = None
        self.os_version = None
        self.os_type = None
        self.family = None
        self.platform = None
        self.udi = None
        # self.is_console = None
        self.prompt = None
        self.prompt_re = None

    def connect(self, ctrl):
        """Connect to the device."""
        if self.prompt:
            self.prompt_re = self.driver.make_dynamic_prompt(self.prompt)
        else:
            self.prompt_re = self.driver.prompt_re

        self.ctrl = ctrl
        if self.protocol.connect(self.driver):
            if self.protocol.authenticate(self.driver):
                self.ctrl.try_read_prompt(1)
                if not self.prompt:
                    self.prompt = self.ctrl.detect_prompt()

                if self.is_target:
                    self.update_config_mode()
                    if self.mode is not None and self.mode != 'global':
                        self.last_error_msg = "Device is not in global mode. Disconnected."
                        self.chain.disconnect()
                        return False

                self.prompt_re = self.driver.make_dynamic_prompt(self.prompt)
                self.connected = True

                if self.is_target is False:
                    if self.os_version is None:
                        self.update_os_version()

                    self.update_hostname()
                else:
                    self._connected_to_target()
                return True

        else:
            self.connected = False
            return False

    def _connected_to_target(self):
        self.update_driver(self.prompt)
        self.after_connect()

        try:
            self.prepare_terminal_session()
        except CommandSyntaxError:
            pass

        if self.os_type is None:
            self.update_os_type()

        self.driver_name = self.os_type

        if self.os_version is None:
            self.update_os_version()

        # delegate to device
        self.prompt_re = self.driver.make_dynamic_prompt(self.prompt)

        # self.prepare_terminal_session()

        if self.udi is None:
            self.update_udi()

        if self.family is None:
            self.update_family()

        if self.platform is None:
            self.update_platform()

        if self.is_console is None:
            self.update_console()

        self.enable(self._get_enable_password())

    def _get_enable_password(self):
        return self.node_info.enable_password if self.node_info.enable_password else self.node_info.password

    def disconnect(self):
        """Disconnect the device."""
        self.chain.connection.log("Disconnecting: {}".format(self))
        if self.connected:
            if self.protocol:
                if self.is_console:
                    while self.mode != 'global':
                        try:
                            self.send('exit', timeout=10)
                        except CommandTimeoutError:
                            break

                self.protocol.disconnect(self.driver)
                self.protocol = None

            self.connected = False
            self.ctrl = None

    def send(self, cmd="", timeout=60, wait_for_string=None, password=False):
        """Send the command to the device and return the output.

        Args:
            cmd (str): Command string for execution. Defaults to empty string.
            timeout (int): Timeout in seconds. Defaults to 60s
            wait_for_string (str): This is optional string that driver
                waits for after command execution. If none the detected
                prompt will be used.
            password (bool): If true cmd representing password is not logged
                and condoor waits for noecho.

        Returns:
            A string containing the command output.

        Raises:
            ConnectionError: General connection error during command execution
            CommandSyntaxError: Command syntax error or unknown command.
            CommandTimeoutError: Timeout during command execution

        """
        if self.connected:
            output = ''
            if password:
                self.chain.connection.log("Sending password")
            else:
                self.chain.connection.log("Sending command: '{}'".format(cmd))

            try:
                output = self.execute_command(cmd, timeout, wait_for_string, password)
            except ConnectionError:
                self.chain.connection.log("Connection lost. Disconnecting.")
                # self.disconnect()
                raise

            if password:
                self.chain.connection.log("Password sent successfully")
            else:
                self.chain.connection.log("Command executed successfully: '{}'".format(cmd))

            return output

        else:
            raise ConnectionError("Device not connected", host=self.hostname)

    def execute_command(self, cmd, timeout, wait_for_string, password):
        """Execute command."""
        try:
            self.last_command_result = None
            self.ctrl.send_command(cmd, password=password)
            if wait_for_string is None:
                wait_for_string = self.prompt_re

            # hide cmd in case it's password for further error messages or exceptions.
            if password:
                cmd = "*** Password ***"

            if not self.driver.wait_for_string(wait_for_string, timeout):
                self.chain.connection.log("Unexpected session disconnect during '{}' command execution".format(cmd))
                raise ConnectionError("Unexpected session disconnect", host=self.hostname)

            if self.last_command_result:
                output = self.last_command_result.replace('\r', '')
            else:
                output = self.ctrl.before.replace('\r', '')

            # not needed. Fixes the issue #11
            # second_line_index = output.find('\n') + 1
            # output = output[second_line_index:]
            return output

        except CommandSyntaxError as e:  # pylint: disable=invalid-name
            self.chain.connection.log("{}: '{}'".format(e.message, cmd))
            e.command = cmd
            # TODO: Verify why lint raises an issue
            raise e  # pylint: disable=raising-bad-type

        except (CommandTimeoutError, pexpect.TIMEOUT):
            self.chain.connection.log("Command timeout: '{}'".format(cmd))
            raise CommandTimeoutError(message="Command timeout", host=self.hostname, command=cmd)

        except ConnectionError as e:  # pylint: disable=invalid-name
            self.chain.connection.log("{}: '{}'".format(e.message, cmd))
            raise

        except pexpect.EOF:
            self.chain.connection.log("Unexpected session disconnect")
            raise ConnectionError("Unexpected session disconnect", host=self.hostname)

        except Exception as e:  # pylint: disable=invalid-name
            self.chain.connection.log("Exception {}".format(e))
            raise ConnectionError(message="Unexpected error", host=self.hostname)

    @property
    def driver_name(self):
        """Return driver name or None."""
        return None if self.driver is None else self.driver.platform

    @driver_name.setter
    def driver_name(self, driver_name):
        if driver_name is None:
            self.chain.connection.log("New driver cannot be None")
            return
        if self.driver is None or driver_name != self.driver.platform:
            self.chain.connection.log('Driver change {} -> {}'.format(self.driver.platform, driver_name))
            self.driver = self.make_driver(driver_name)
            self.make_dynamic_prompt(self.prompt)
        else:
            self.chain.connection.log('Driver {}'.format(driver_name))

    def make_driver(self, driver_name='generic'):
        """Make driver factory function."""
        module_str = 'condoor.drivers.%s' % driver_name
        try:
            __import__(module_str)
            module = sys.modules[module_str]
            driver_class = getattr(module, 'Driver')
        except ImportError as e:  # pylint: disable=invalid-name
            self.chain.connection.log("Import error", exc_info=e)
            return self.make_driver()
            # raise GeneralError("Platform {} not supported".format(driver_name))

            self.chain.connection.log("Make Device: {} with Driver: {}".format(self, driver_class.platform))
        return driver_class(self)

    def get_previous_prompts(self):
        """Return list of prompts from all devices except target."""
        return self.chain.get_previous_prompts(self)

    @property
    def version_text(self):
        """Return version text and collect if not available."""
        if self._version_text is None:
            self.chain.connection.log("Collecting version information")
            self._version_text = self.driver.get_version_text()
            if self._version_text:
                self.chain.connection.log("Version info collected")
            else:
                self.chain.connection.log("Version info not collected")

        return self._version_text

    @property
    def hostname_text(self):
        """Return hostname text and collect if not available."""
        if self._hostname_text is None:
            self.chain.connection.log("Collecting hostname information")
            self._hostname_text = self.driver.get_hostname_text()
            if self._hostname_text:
                self.chain.connection.log("Hostname info collected")
            else:
                self.chain.connection.log("Hostname info not collected")
        return self._hostname_text

    @property
    def inventory_text(self):
        """Return inventory information and collect if not available."""
        if self._inventory_text is None:
            self.chain.connection.log("Collecting inventory information")
            self._inventory_text = self.driver.get_inventory_text()
            if self._inventory_text:
                self.chain.connection.log("Inventory info collected")
            else:
                self.chain.connection.log("Inventory info not collected")
        return self._inventory_text

    @property
    def users_text(self):
        """Return connected users information and collect if not available."""
        if self._users_text is None:
            self.chain.connection.log("Getting connected users text")
            self._users_text = self.driver.get_users_text()
            if self._users_text:
                self.chain.connection.log("Users text collected")
            else:
                self.chain.connection.log("Users text not collected")
        return self._users_text

    def get_protocol_name(self):
        """Provide protocol name based on node_info."""
        protocol_name = self.node_info.protocol
        if self.is_console:
            protocol_name += '_console'
        return protocol_name

    def make_dynamic_prompt(self, prompt):
        """Extend prompt with flexible mode handling regexp."""
        if prompt:
            self.prompt_re = self.driver.make_dynamic_prompt(prompt)

    def update_udi(self):
        """Update udi."""
        self.chain.connection.log("Parsing inventory")
        # TODO: Maybe validate if udi is complete
        self.udi = parse_inventory(self.inventory_text)

    def update_config_mode(self, prompt=None):
        """Update config mode."""
        # TODO: Fix the conflict with config mode attribute at connection
        if prompt:
            self.mode = self.driver.update_config_mode(prompt)
        else:
            self.mode = self.driver.update_config_mode(self.prompt)

    def update_hostname(self):
        """Update hostname."""
        self.hostname = self.driver.update_hostname(self.prompt)

    def update_driver(self, prompt):
        """Update driver based on new prompt."""
        prompt = prompt.lstrip()
        self.chain.connection.log("({}): Prompt: '{}'".format(self.driver.platform, prompt))
        self.prompt = prompt
        driver_name = self.driver.update_driver(prompt)
        if driver_name is None:
            self.chain.connection.log("New driver not detected. Using existing {} driver.".format(self.driver.platform))
            return
        self.driver_name = driver_name

    def prepare_terminal_session(self):
        """Send commands to prepare terminal session configuration."""
        for cmd in self.driver.prepare_terminal_session:
            self.send(cmd)

    def update_os_type(self):
        """Update os_type attribute."""
        os_type = self.driver.get_os_type(self.version_text)
        if os_type:
            self.chain.connection.log("SW Type: {}".format(os_type))
            self.os_type = os_type

    def update_os_version(self):
        """Update os_version attribute."""
        os_version = self.driver.get_os_version(self.version_text)
        if os_version:
            self.chain.connection.log("SW Version: {}".format(os_version))
            self.os_version = os_version

    def update_family(self):
        """Update family attribute."""
        family = self.driver.get_hw_family(self.version_text)
        if family:
            self.chain.connection.log("HW Family: {}".format(family))
            self.family = family

    def update_platform(self):
        """Update platform attribute."""
        platform = self.driver.get_hw_platform(self.udi)
        if platform:
            self.chain.connection.log("HW Platform: {}".format(platform))
            self.platform = platform

    def update_console(self):
        """Update is_console whether connected via console."""
        is_console = self.driver.is_console(self.users_text)
        if is_console is not None:
            self.is_console = is_console

    def after_connect(self):
        """Execute right after connect."""
        return self.driver.after_connect()

    def enable(self, enable_password):
        """Set privilege mode."""
        self.driver.enable(enable_password)

    def reload(self, reload_timeout, save_config, no_reload_cmd):
        """Reload device."""
        if not no_reload_cmd:
            self.ctrl.send_command(self.driver.reload_cmd)
        return self.driver.reload(reload_timeout, save_config)

    def run_fsm(self, name, command, events, transitions, timeout, max_transitions=20):
        """Wrap the FSM code."""
        self.ctrl.send_command(command)
        return FSM(name, self, events, transitions, timeout=timeout, max_transitions=max_transitions).run()

    def config(self, configlet, plane, **attributes):
        """Apply config to the device."""
        try:
            config_text = configlet.format(**attributes)
        except KeyError as exp:
            raise CommandSyntaxError("Configuration template error: {}".format(str(exp)))

        return self.driver.config(config_text, plane)

    def rollback(self, label=None, plane='sdr'):
        """Rollback config on the device."""
        return self.driver.rollback(label, plane)



"""Provides SSH driver class."""

from functools import partial
import pexpect

from condoor.fsm import FSM, action
from condoor.utils import pattern_to_str
from condoor.protocols.base import Protocol
from condoor.actions import a_send_password, a_authentication_error, a_send, a_unable_to_connect, a_save_last_pattern,\
    a_send_line

from condoor.exceptions import ConnectionError, ConnectionTimeoutError
from condoor.config import CONF


MODULUS_TOO_SMALL = "modulus too small"
PROTOCOL_DIFFER = "Protocol major versions differ"
NEWSSHKEY = "fingerprint is"
KNOWN_HOSTS = "added.*to the list of known hosts"
HOST_KEY_FAILED = "key verification failed"

_C = CONF['protocol']['ssh']


class SSH(Protocol):
    """SSH protocol implementation."""

    def __init__(self, device):
        """Initialize SSH object."""
        super(SSH, self).__init__(device)

    def get_command(self, version=2):
        """Return the SSH protocol specific command to connect."""
        try:
            options = _C['options']
            options_str = " -o ".join(options)
            if options_str:
                options_str = "-o " + options_str + " "
        except KeyError:
            options_str = ""

        if self.username:
            # Not supported on SunOS
            # "-o ConnectTimeout={}
            command = "ssh {}" \
                      "-{} " \
                      "-p {} {}@{}".format(options_str, version, self.port, self.username, self.hostname)
        else:
            command = "ssh {} " \
                      "-{} " \
                      "-p {} {}".format(options_str, version, self.port, self.hostname)
        return command

    def connect(self, driver):
        """Connect using the SSH protocol specific FSM."""
        #                      0                    1                 2
        events = [driver.password_re, self.device.prompt_re, driver.unable_to_connect_re,
                  #   3          4              5               6                   7
                  NEWSSHKEY, KNOWN_HOSTS, HOST_KEY_FAILED, MODULUS_TOO_SMALL, PROTOCOL_DIFFER,
                  #      8              9
                  driver.timeout_re, pexpect.TIMEOUT]

        transitions = [
            (driver.password_re, [0, 1, 4, 5], -1, partial(a_save_last_pattern, self), 0),
            (self.device.prompt_re, [0], -1, partial(a_save_last_pattern, self), 0),
            #  cover all messages indicating that connection was not set up
            (driver.unable_to_connect_re, [0], -1, a_unable_to_connect, 0),
            (NEWSSHKEY, [0], 1, partial(a_send_line, "yes"), 10),
            (KNOWN_HOSTS, [0, 1], 0, None, 0),
            (HOST_KEY_FAILED, [0], -1, ConnectionError("Host key failed", self.hostname), 0),
            (MODULUS_TOO_SMALL, [0], 0, self.fallback_to_sshv1, 0),
            (PROTOCOL_DIFFER, [0], 4, self.fallback_to_sshv1, 0),
            (PROTOCOL_DIFFER, [4], -1, ConnectionError("Protocol version differs", self.hostname), 0),
            (pexpect.TIMEOUT, [0], 5, partial(a_send, "\r\n"), 10),
            (pexpect.TIMEOUT, [5], -1, ConnectionTimeoutError("Connection timeout", self.hostname), 0),
            (driver.timeout_re, [0], -1, ConnectionTimeoutError("Connection timeout", self.hostname), 0),
        ]

        self.log("EXPECTED_PROMPT={}".format(pattern_to_str(self.device.prompt_re)))
        fsm = FSM("SSH-CONNECT", self.device, events, transitions, timeout=_C['connect_timeout'],
                  searchwindowsize=160)
        return fsm.run()

    def authenticate(self, driver):
        """Authenticate using the SSH protocol specific FSM."""
        #              0                     1                    2                  3
        events = [driver.press_return_re, driver.password_re, self.device.prompt_re, pexpect.TIMEOUT]

        transitions = [
            (driver.press_return_re, [0, 1], 1, partial(a_send, "\r\n"), 10),
            (driver.password_re, [0], 1, partial(a_send_password, self._acquire_password()),
             _C['first_prompt_timeout']),
            (driver.password_re, [1], -1, a_authentication_error, 0),
            (self.device.prompt_re, [0, 1], -1, None, 0),
            (pexpect.TIMEOUT, [1], -1,
             ConnectionError("Error getting device prompt") if self.device.is_target else partial(a_send, "\r\n"), 0)
        ]

        self.log("EXPECTED_PROMPT={}".format(pattern_to_str(self.device.prompt_re)))
        fsm = FSM("SSH-AUTH", self.device, events, transitions, init_pattern=self.last_pattern, timeout=30)
        return fsm.run()

    def disconnect(self, driver):
        """Disconnect using the protocol specific method."""
        self.log("SSH disconnect")
        try:
            self.device.ctrl.sendline('\x03')
            self.device.ctrl.sendline('\x04')
        except OSError:
            self.log("Protocol already disconnected")

    # FIXME: This needs to be fixed and tested
    @action
    def fallback_to_sshv1(self, ctx):
        """Fallback to SSHv1."""
        command = self.get_command(version=1)
        ctx.spawn_session(command)
        return True



"""Provides the PatternManager class."""

import os
import re
from utils import yaml_file_to_dict


class PatternManager(object):
    """Provides API to patterns defined externally."""

    def __init__(self, pattern_dict):
        """Initialize PatternManager object."""
        self._dict = pattern_dict
        self._dict_compiled, self._dict_text, self._dict_dscr = self._prepare_patterns(pattern_dict)

    def _prepare_patterns(self, pattern_dict):
        """Return two dictionaries: compiled and text prompts."""
        dict_compiled = {}
        dict_text = {}
        dict_dscr = {}
        for platform, patterns in pattern_dict.items():
            dict_text[platform] = {}
            dict_compiled[platform] = {}
            dict_dscr[platform] = {}

            for key, pattern in patterns.items():
                try:
                    text_pattern = None
                    compiled_pattern = None
                    description_pattern = None

                    if isinstance(pattern, str):
                        text_pattern = pattern
                        compiled_pattern = re.compile(text_pattern, re.MULTILINE)
                        description_pattern = key

                    elif isinstance(pattern, dict):
                        text_pattern = pattern['pattern']
                        compiled_pattern = re.compile(text_pattern, re.MULTILINE)
                        description_pattern = pattern['description']

                    elif isinstance(pattern, list):
                        text_pattern = self._concatenate_patterns(key, pattern)
                        compiled_pattern = re.compile(text_pattern, re.MULTILINE)
                        description_pattern = key

                    dict_text[platform][key] = text_pattern
                    dict_compiled[platform][key] = compiled_pattern
                    dict_dscr[platform][key] = description_pattern

                except re.error as e:
                    raise RuntimeError("Pattern compile error: {} ({}:{})".format(e.message, platform, key))

        return dict_compiled, dict_text, dict_dscr

    def _platform_patterns(self, platform='generic', compiled=False):
        """Return all the patterns for specific platform."""
        patterns = self._dict_compiled.get(platform, None) if compiled else self._dict_text.get(platform, None)
        if patterns is None:
            raise KeyError("Unknown platform: {}".format(platform))
        return patterns

    def _concatenate_patterns(self, key, patterns):
        pattern_set = set()
        for platform in patterns:
            try:
                pattern = self._dict.get(platform, None)[key]
                if isinstance(pattern, dict):
                    pattern = pattern.get('pattern', None)

                pattern_set |= set(pattern.split('|'))
            except (KeyError, TypeError):
                continue
        return "|".join(pattern_set)

    def pattern(self, platform, key, compiled=True):
        """Return the pattern defined by the key string specific to the platform.

        :param platform:
        :param key:
        :param compiled:
        :return: Pattern string or RE object.
        """
        patterns = self._platform_patterns(platform, compiled=compiled)
        pattern = patterns.get(key, self._platform_patterns(compiled=compiled).get(key, None))

        if pattern is None:
            raise KeyError("Patterns database corrupted. Platform: {}, Key: {}".format(platform, key))

        return pattern

    def description(self, platform, key):
        """Return the patter description."""
        patterns = self._dict_dscr.get(platform, None)
        description = patterns.get(key, None)
        return description

    def platform(self, with_prompt, platforms=None):
        """Return the platform name based on the prompt matching."""
        if platforms is None:
            platforms = self._dict['generic']['prompt_detection']

        for platform in platforms:
            pattern = self.pattern(platform, 'prompt')
            result = re.search(pattern, with_prompt)
            if result:
                return platform
        return None


class YPatternManager(PatternManager):
    """Yaml version of pattern manager."""

    def __init__(self):
        """Initialize the pattern manager object."""
        script_name = os.path.splitext(__file__)[0]
        path = os.path.abspath('./')
        super(YPatternManager, self).__init__(pattern_dict=yaml_file_to_dict(script_name, path))

# ypm = YPatternManager()
# from pprint import pprint
# pprint(ypm._dict_compiled)
# pprint(ypm._dict_text)
# pprint(ypm.pattern('eXR', 'prompt'))
# pprint(ypm.pattern('eXR', 'prompt', compiled=False))
#
# pprint(ypm.pattern('eXR', 'press_return'))
# pprint(ypm.pattern('eXR', 'press_return', compiled=False))
#
# pattern = ypm.pattern('eXR', 'press_return')
# print(re.search(pattern, "Press RETURN to get started."))
#
# description = ypm.description('eXR', 'syntax_error')
# print(description)

# p = ypm.get_pattern('generic', "syntax_error")
# pprint(p.pattern)

# print(ypm.get_platform_based_on_prompt('[sysadmin-vm:0_RSP0:~]$'))
# print(ypm.get_platform_based_on_prompt('sysadmin-vm:0_RSP0#'))

# print(ypm.get_pattern("generic", "syntax_error", compiled=False))
# print(ypm.get_pattern("generic", "syntax_error").pattern)

# print(ypm._get_all_patterns('rommon').pattern)

# print(ypm._get_all_patterns('rommon', compiled=False))

# print(ypm._get_all_patterns('dupa'))

# print(ypm.get_pattern("XR", "connection_closed", compiled=False))
# print(ypm.get_pattern("XR", "standby_console", compiled=False))



"""This is a Calvados driver implementation."""

import re
from condoor.drivers.generic import Driver as Generic
from condoor import pattern_manager


class Driver(Generic):
    """This is a Driver class implementation for Calvados."""

    platform = 'Calvados'
    inventory_cmd = 'show inventory chassis'
    target_prompt_components = ['prompt_dynamic', 'prompt_default', 'exr', 'windriver']
    prepare_terminal_session = ['terminal len 0', 'terminal width 0']
    families = {
        "ASR9K": "ASR9K",
        "ASR-9": "ASR9K",
        "ASR9": "ASR9K",
        "NCS-6": "NCS6K",
        "NCS-4": "NCS4K",
        "NCS-50": "NCS5K",
        "NCS-55": "NCS5500",
        "NCS1": "NCS1K",
        "NCS-1": "NCS1K",
    }

    def __init__(self, device):
        """Initialize Calvados driver object."""
        super(Driver, self).__init__(device)

    def get_version_text(self):
        """Return the version information."""
        version_text = self.device.send('show version', timeout=60)
        return version_text

    def update_driver(self, prompt):
        """Return driver name based on prompt analysis."""
        return pattern_manager.platform(prompt, ['Calvados', 'Windriver', 'eXR'])

    def after_connect(self):
        """Execute after connect."""
        # TODO: check if this works.
        show_users = self.device.send("show users", timeout=120)
        result = re.search(pattern_manager.pattern(self.platform, 'connected_locally'), show_users)
        if result:
            self.log('Locally connected to Calvados. Exiting.')
            self.device.send('exit')
            return True
        return False
